# Copyright 2022 Sony Semiconductors Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from model_compression_toolkit.common.target_platform.targetplatform2framework.current_tpc import get_current_tpc
from model_compression_toolkit.common.target_platform.targetplatform2framework.target_platform_capabilities import TargetPlatformCapabilities
from model_compression_toolkit.common.target_platform.targetplatform2framework.attribute_filter import \
    Eq, GreaterEq, NotEq, SmallerEq, Greater, Smaller
from model_compression_toolkit.common.target_platform.targetplatform2framework.layer_filter_params import \
    LayerFilterParams
from model_compression_toolkit.common.target_platform.targetplatform2framework.operations_to_layers import \
    OperationsToLayers, OperationsSetToLayers





